/*
 * 2004 Viliam Holub
 */


/*
 * Include standard headers
 */
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>

#include "png_basics.h"


// color definitions
color_s white	= { 0xff, 0xff, 0xff, 0xff};
color_s gray	= { 0x80, 0x80, 0x80, 0xff};
color_s black	= { 0x00, 0x00, 0x00, 0xff};
color_s red	= { 0xff, 0x00, 0x00, 0xff};
color_s green	= { 0x00, 0xff, 0x00, 0xff};
color_s blue	= { 0x00, 0x00, 0xff, 0xff};


const char *png_filename;


/** Shows error message on the output and exits.
 *
 * die is used to terminate the program when an error occurs.
 */
static void
die( const char *fmt, ...)

{
	va_list ap;
	
	fflush( stdout);
	
	fprintf( stderr, "error: ");
	va_start( ap, fmt);
	vfprintf( stderr, fmt, ap);
	va_end( ap);
	
	fprintf( stderr, "\n");
	exit( -1);
}


/** Allocates a memory block.
 *
 * The allocation will not fail. In case there is not enough memory
 * the error message is displayed and the program is aborted.
 */
static void *
xmalloc( size_t size)

{
	void *v = malloc( size);
	if (!v)
		die( "Out of memory");
	return v;
}


/*
 * PNG error reporting functions.
 *
 */

/** Dies with png error message.
 */
void
png_error_handler( png_structp png, const char *err_str)

{
	die( "%s: %s", png_filename, err_str);
}


/** Shows png error message but does not halt the program.
 */
void
png_warning_handler( png_structp png, const char *err_str)

{
	fprintf( stderr, "error: %s: warning - %s", png_filename,
			err_str);
}



/****************************************************************************
 * Picture structure manipulating functions
 *
 *
 */


/** Allocs a new picture and initializes its attributes.
 */
picture_s *
new_picture( void)

{
	picture_s *pic = (picture_s *)xmalloc( sizeof( picture_s));

	pic->rows = NULL;
	pic->width = -1;
	pic->height = -1;

	return pic;
}


/** Sets picture attributes according to the width and height.
 */
void
init_picture( picture_s *pic, int width, int height)

{
	int row;

	pic->width = width;
	pic->height = height;
	pic->rows = (png_byte **)malloc( sizeof( png_bytep) *height);
	for (row=0; row<height; row++)
	{
		pic->rows[ row] = (png_byte *)malloc( width*4);
		memset( pic->rows[ row], 0, width*4);
	}
}


/** Free all data structures.
 */
void
done_picture( picture_s **pic)

{
	int row;

	for (row=0; row<(*pic)->height; row++)
		free( (*pic)->rows[ row]);
	free( (*pic)->rows);
	free( *pic);
	*pic = NULL;
}


/** Uploads RGB PNG file into picture structure.
 */
picture_s *
load_rgbpicture( const char *filename)

{
	int row;
	FILE *src;
	png_infop pngs_info;
	png_structp pngs;
	int color_type;
	int row_bytes;
	picture_s *pic = new_picture();
	
	png_filename = filename;

	src = fopen( filename, "rb");
	if (!src)
		die( "%s: %s", filename, strerror( errno));

	pngs = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL,
			png_error_handler,
			png_warning_handler);

	pngs_info = png_create_info_struct( pngs);
	png_init_io( pngs, src);
	png_read_info( pngs, pngs_info);
	color_type = png_get_color_type( pngs, pngs_info);
	png_set_strip_16( pngs); 
	if (color_type == PNG_COLOR_TYPE_GRAY ||
			color_type == PNG_COLOR_TYPE_GA)
		png_set_gray_to_rgb( pngs);
	if (color_type == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb( pngs);
	png_set_filler( pngs, 0xff, PNG_FILLER_AFTER);
	png_read_update_info( pngs, pngs_info);

	pic->height = png_get_image_height( pngs, pngs_info);
	pic->width  = png_get_image_width( pngs, pngs_info);

	pic->rows = (png_byte **)xmalloc( sizeof( png_bytep) *pic->height);

	row_bytes = png_get_rowbytes( pngs, pngs_info);
	for (row = 0; row < pic->height; row++)
		pic->rows[ row] = (png_byte *)xmalloc( row_bytes);

	png_read_image( pngs, pic->rows);
	png_read_end( pngs, pngs_info);

	png_destroy_read_struct( &pngs, &pngs_info, png_infopp_NULL);

	fclose( src);

	return pic;
}


/** Saves the picture into the file.
 */
void
save_picture( picture_s *pic, const char *filename)

{
	FILE *fd;
	png_infop pngs_info;
	png_structp pngs;

	fd = fopen( filename, "wb");
	if (!fd)
		die( "%s: %s", filename, strerror( errno));

	png_filename = filename;
	pngs = png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL,
			png_error_handler,
			png_warning_handler);
	pngs_info = png_create_info_struct( pngs);
	png_init_io( pngs, fd);
	png_set_compression_level( pngs, Z_BEST_SPEED);
	png_set_IHDR( pngs, pngs_info, pic->width, pic->height, 8,
			PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_set_rows( pngs, pngs_info, pic->rows);
	png_write_png( pngs, pngs_info, 0, NULL);
	png_destroy_write_struct( &pngs, &pngs_info);

	fclose( fd);
}




/****************************************************************************
 * Image manipulating functions
 *
 * 
 */


/** Gets the pixel color.
 */
void
get_pixel( const picture_s *pic, color_s *col, int x, int y)

{
	/*
	if (x < 0 || x >= pic->width || y < 0 || y >= pic->height)
	{
		*col = white;
		return;
	}
	*/

	x *= 4;
	col->r = pic->rows[ y][ x+0];
	col->g = pic->rows[ y][ x+1];
	col->b = pic->rows[ y][ x+2];
	col->a = pic->rows[ y][ x+3];
}


/** Writes the picture color.
 */
void
put_pixel( picture_s *pic, int x, int y, color_s *col)

{
	if (x < 0 || x >= pic->width || y < 0 || y >= pic->height ||
			col->a == 0x00)
		return;

	x *= 4;
	pic->rows[ y][ x+0] = col->r;
	pic->rows[ y][ x+1] = col->g;
	pic->rows[ y][ x+2] = col->b;
	pic->rows[ y][ x+3] = col->a;
}


/** Fills the line with specified color.
 */
void
put_line( picture_s *pic, int x, int y, int c, color_s *col)

{
	if (c <= 0 || x < 0 || x+c > pic->width || y < 0 ||
			y >= pic->height || col->a == 0x00)
		return;

	x *= 4;

	for (; c; c--, x+=4)
	{
		pic->rows[ y][ x+0] = col->r;
		pic->rows[ y][ x+1] = col->g;
		pic->rows[ y][ x+2] = col->b;
		pic->rows[ y][ x+3] = col->a;
	}
}
